/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

/* configurator_xml.c: an xml configurator */

/* Developed using:
 * ii  libxml2                   2.3.9-0.1                 GNOME XML library
 * ii  libxml2-dev               2.3.9-0.1                 Development files for the GNOME XML library
 */

static char rcsid[] = "$Id: configurator_xml.c,v 1.8 2001/11/10 01:45:33 tlewis Exp $";

#include "hank.h"
#include "configuration.h"
#include "configurator.h"
#include "matcher.h"
#include "memcon.h"
#include "protocol_engine.h"
#include "protocol_lookup.h"
#include "support.h"
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <stdlib.h>

extern char *strdup(char*);

static xmlDocPtr doc;

extern int xmlDoValidityCheckingDefaultValue;

extern struct protocol_lookup pls;
extern lt proto_table;

extern struct configuration *configuration;
extern struct memcon *mc;
extern struct event_handler mtl;

static int prid=0;
static int rid=0;

static struct memcon_context *mcon;

static struct configuration* my_configure(int argc, char **argv);
static int my_release_configuration(struct configuration *c);
static void my_do_stuff(int signum);

struct configurator configurator_xml = {
	{
		"configurator",
		"xml",
		1,
		1,
		&configurator_xml
	},
	&my_configure,
	&my_release_configuration,
	&my_do_stuff
};

#define STATIC_MODULE_NAME xml

#ifdef ENABLE_DYN_MODULES
struct smodule * paengine [2] = {
#else  /*ENABLE_DYN_MODULES*/
struct smodule * STATIC_MODULE_NAME [2] =  {
#endif /*ENABLE_DYN_MODULES*/
	&(configurator_xml.s),
	NULL
};

static struct melement*
build_rule(struct melement *parent, xmlNodePtr x)
{
	struct melement *m;
	xmlNodePtr y;

	/* recursion rocks */
	x=get_real_xml_node(x);
	if(x&&!strcmp((const char *)x->name, "AND")){
		if(m=mtl.add_element(parent, ETYPE_AND))
			for(y=x->CHILDREN;y;y=y->NEXT)
				if(!TYPE==TEXT)
					build_rule(m, y);
	} else if(x&&!strcmp((const char *)x->name, "OR")){
		if(m=mtl.add_element(parent, ETYPE_OR))
			for(y=x->CHILDREN;y;y=y->NEXT)
				if(!TYPE==TEXT)
					build_rule(m, y);
	} else if(x&&!strcmp((const char *)x->name, "PRIM")) {
		char *proto, *section, *field, *op, *arg, *not;

		proto=(char *)xmlGetProp(x, (unsigned const char *)"proto");
		section=(char *)xmlGetProp(x, (unsigned const char *)"section");
		field=(char *)xmlGetProp(x, (unsigned const char *)"field");
		op=(char *)xmlGetProp(x, (unsigned const char *)"op");
		not=(char *)xmlGetProp(x, (unsigned const char *)"not");
		arg=(char *)x->VALUE;

		m=mtl.add_primitive(parent, PROTO, SECTION, FIELD, OP, ARG, NOT);
	} else {
		/* log; error. */
		return(NULL);
	}

	return(m);
}

static void
process_prim(xmlNodePtr x)
{
	/* FIXME: this function is non-sensical and needs to be made real */
	char *eng, *section, *field,  *op, *arg;

	eng=(char *)xmlGetProp(x, (unsigned const char *)"eng");
	if((r->content.p.pe=pls.find_pe_name(proto_table, eng))==NULL){
		printf("PE lookup on %s failed.\n", eng);
		exit(-1);
	} else {
		printf("PE lookup on %s (%s)  succeeded!\n", eng, r->content.p.pe->s.name);
	}

	arg=strdup((char *)x->children->content);

	r->disc=CTYPE_PRIM;
	r->comment=NULL;
	r->content.p.next=NULL;
	r->content.p.arg=strdup(arg);
	/* printf("Registering primitive %i: op=%s, arg=%s\n", prid, op, arg); */
	if((prid=r->content.p.pe->reg_prim(mcon, op, arg))==-1){
		printf("Registration of primitive failed!  Aborting.\n");
		exit(-1);
	} else {
		r->content.p.pid=prid;
		prid++;
	}

	if(r->content.p.pid==-1){
		printf("REG_PRIM: ERROR! %s-%s: %s -- returned %i\n", 
			r->content.p.pe->s.name, 
			op, arg, r->content.p.pid);
		return(NULL);
	}
	printf("REG_PRIM: success. name=%s, op=%s, arg=%s -- returned %i\n", 
		r->content.p.pe->s.name, 
		op, arg, r->content.p.pid);
	free(arg);
}

static void
process_rule(xmlNodePtr x)
{
	struct rule *r;
	int i;

	while(!strcmp((const char *)x->name, "text"))
		x=x->next;

	if(!strcmp((const char *)x->name, "rule")){
		printf("Got rule!\n");
		x=x->children;
		while(x){
			if(!strcmp((const char *)x->name, "match"))
				break;
			x=get_real_xml_node(x->next);
		}

		if(x){
			x=x->children;
			while((x=x->next)){
				if(((r=build_rule(x))!=NULL)){
					break;
				}
			}
		} else {
			printf("No joy on this rule.  Shucks.\n");
			return;
		}
	}
	printf("REG_RULE with %i; ", rid);
	r->rid=rid;
	i=mtl.reg_rule(r, rid);
	rid++;
	printf("returned %i.\n", i);
}

static struct configuration* 
my_configure(int argc, char **argv)
{
	struct configuration *c=NULL;
	xmlValidCtxt v;
	int i;
	static xmlNodePtr cur, x;

	mcon=mc->create(NULL, "configurator-xml");

	xmlDoValidityCheckingDefaultValue=1;

	if(argc!=1)
		return(NULL);

	mtl.configure(NULL);

	if((c=malloc(sizeof(struct configuration)))==NULL)
		return(NULL);
	memset(c, 0, sizeof(*c));

	doc = xmlParseFile(argv[0]);
	if(doc==NULL)
		return(NULL);

	if(xmlValidateDocument(&v, doc)!=1){
		fprintf(stderr, "error: xmlValidateDocument() returned %i.\n", i);
		fprintf(stderr, "It appears that the file %s is not a valid config file.\n", argv[0]);
		return(NULL);
	} else {
		/* fprintf(stderr, "Document validated properly.\n"); */
	}

	cur = xmlDocGetRootElement(doc);
	if (cur == NULL) {
		fprintf(stderr,"empty document\n");
		xmlFreeDoc(doc);
		return(NULL);
	}

	if (strcmp((const char*)cur->name, "hank")){
		fprintf(stderr, "Root is not of type \"hank\"!\n");
		return(NULL);
	}

	cur=cur->children;

	printf("Looking for rules: ");
	while(cur!=NULL){
		if (!strcmp((const char*)cur->name, "rules")){
			x=cur->children;
			while(x!=NULL){
				if (!strcmp((const char*)x->name, "rule")){
					fprintf(stderr,"Processing rule:\n");
					process_rule(x);
				}
				x=x->next;
			}
		}
		cur=cur->next;
	}

	return(NULL);
}

static struct rule* 
my_load_rules(void)
{
	return(NULL);
}

static int 
my_release_rules(struct rule *rules)
{
	if(rules==NULL)
		return(1);
	return(0);
}

static int 
my_release_configuration(struct configuration *c)
{
	if(c==NULL)
		return(1);
	return(0);
}

static void 
my_do_stuff(int signum)
{
	signum++;
}
