/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

#include "support.h"
#include "event.h"
#include "packet.h"
#include <libxml/parser.h>

static char rcsid[] = "$Id: event.c,v 1.21 2001/12/08 00:04:48 tlewis Exp $";

extern event_id suigeneris;
extern struct event_handler matcher;
extern struct event_handler reporter;
extern struct event_handler firewall;
extern int xmlDoValidityCheckingDefaultValue;

#define PRINTF(x...) 

struct collection_elem {
	event_id id;
	struct collection_elem *next;
};

static struct collection_elem *collections[16384];
static int ncols = 16383;

static event_id
create_collection(void)
{
	int i;

	if((i=ncols--)<1){
		fprintf(stderr, "AIEE!  Too many rule collections.  Aborting.\n");
		exit(-1);
	}
	return( (EH_GROUP<<24) | i );
}

static void
handle_collection_event(struct packet *p)
{
	struct event *e, *f, *g;
	struct collection_elem *c;

	for(e=p->q;e;e=e->next){
		if((e->id>>24)==EH_GROUP){
			for(c=collections[(e->id&0xffffff)];c;c=c->next){
				if(((f=p->c->eng->alloc(p->c, sizeof(*f))))==NULL){
					fprintf(stderr, "AIEE!  malloc() failed.  Aborting.\n");
					exit(-1);
				}
				f->id=c->id;
				f->consumed=0;
				f->payload=NULL;
				f->next=NULL;
				for(g=p->q;g->next;g=g->next)
					;
				g->next=f;
			}
			e->consumed=1;
		}
	}
}

static void
register_collection(event_id id, xmlNodePtr node)
{
	int i = (id & 0xffffff);
	struct collection_elem *e;
	xmlNodePtr x;

	for(x=node->children;x;x=x->next){

		if((e=malloc(sizeof(*e)))==NULL) {
			fprintf(stderr, "AIEE!  malloc failed.  Aborting.\n");
			exit(-1);
		}

		if( (!strcmp((char *)x->name, "text")) || (!strcmp((char *)x->name, "comment")) ){
			/* skip */
			free(e);
		} else {
			if(!strcmp((char *)x->name, "report")){
				e->id=reporter.create_event();
				if(reporter.register_handler(e->id, 0, x)){
					printf("Error registering report; exiting.\n");
					exit(-1);
				}
			} else if(!strcmp((char *)x->name, "firewall")){
				e->id=firewall.create_event();
				if(firewall.register_handler(e->id, 0, x)){
					printf("Error registering firewall; exiting.\n");
					exit(-1);
				}
			} else {
				printf("ERROR: unrecognized keyword \"%s\"; skipping.\n", (char *)x->name);
			}
			e->next=collections[i];
			collections[i]=e;
		}

	}

}

/* private functions */
static void
establish_rule(xmlNodePtr rule, event_id incoming)
{
	event_id lead_in=0;
	xmlNodePtr x, y;

	/* We here thank God and Daniel Veillard for dtd validation. */
	for(x=rule->children;x;x=x->next){
		if(!strcmp((char *)x->name, "text")){
			/* skip */
		} else if(!strcmp((char *)x->name, "comment")){
			/* skip */
		} else if(!strcmp((char *)x->name, "match")){
			for(y=x->children;y;y=y->next)
				if ( (strcmp((char *)y->name, "text")) && (strcmp((char *)y->name, "comment")) )
					if(matcher.register_handler(incoming, lead_in, y)){
						printf("Error registering match; exiting.\n");
						exit(-1);
					}
		} else if(!strcmp((char *)x->name, "collection")){
			lead_in=create_collection();
			register_collection(lead_in, x);
		} else {
			printf("ERROR: unrecognized keyword \"%s\"; skipping.\n", (char *)x->name);
		}
	}
}

/* public functions */

/* Ok, for now I don't honor any call-once targets.  When I do, then
 * I will need to track precedence for all handlers and, when a new
 * handler is registered, check to ensure that it creates no precedence
 * inversions. */

/* decomposition, matching and reporting are the main targets right now.
 * The decomp target needs to be registered with the paengine as the
 * target at which to point a natal packet. */
void
establish_rules(const char *xml_rule_file)
{
	event_id match_collection;
	xmlDocPtr doc;
	xmlNodePtr x, y;
	xmlValidCtxt v;
	int i;

	/* xmlDoValidityCheckingDefaultValue=1; */
	xmlSubstituteEntitiesDefault(1);
	xmlLineNumbersDefault(1);

	if((doc = xmlParseFile(xml_rule_file))==NULL){
		fprintf(stderr, "Could not open rule file %s; exiting.\n", xml_rule_file);
		exit(-1);
	}

/*
	if((i=xmlValidateDocument(&v, doc))!=1){
		fprintf(stderr, "error: xmlValidateDocument() returned %i.\n", i);
		fprintf(stderr, "It appears that the file %s is not a valid rules file.\n", xml_rule_file);
		fprintf(stderr, "Exiting.\n");
		exit(-1);
	}
*/

	x = xmlDocGetRootElement(doc);

	/* create decomp target */
	suigeneris = matcher.create_event();

	/* create match target */
	match_collection=matcher.create_event();

	for(x=x->children;x;x=x->next){
		int nr=1;
		if(!strcmp((char *)x->name, "rules")){
			for(y=x->children;y;y=y->next){
				if(!strcmp((char *)y->name, "rule")){
					/* printf("Registering rule %i\n", nr++); */
					establish_rule(y, match_collection);
				} else {
					/* printf("Skipping \"%s\" from event...\n", y->name); */
				}
			}
		} else if(!strcmp((char *)x->name, "decomp")){
			PRINTF("Registering decomp from event...\n");
			matcher.register_handler(suigeneris, match_collection, x);
		} else {
			/* printf("Skipping \"%s\" from event...\n", x->name); */
		}
	}

}

void
handle_events(struct packet *p)
{
	struct event *e;
	struct event_handler *h;

	/* this is emminently cpp'able */
	for(e=p->q;e;e=e->next){
		unsigned long ls, le;
		rdtscl(ls);
		if(!e->consumed){
			if((e->id==0)){
				/* skip */
			} else if((e->id>>24)==EH_MATCHER){
				matcher.handle_event(p);
			} else if((e->id>>24)==EH_REPORTER){
				reporter.handle_event(p);
			} else if((e->id>>24)==EH_FIREWALL){
				firewall.handle_event(p);
			} else if((e->id>>24)==EH_GROUP){
				handle_collection_event(p);
			} else {
				printf("UNKNOWN EVENT: (%i, %i.%i); skipping.\n", e->id, e->id>>24, e->id&0xffffff);
			}
		} else {
			/* printf(" Skipping already-run event %i.%i (%i).\n", \
				e->id>>24, e->id&0xffffff, e->consumed); */
		}
		rdtscl(le);
		printf("Handling event of class %i took %i cycles.\n", 
			e->id>>24, (le-ls));
	}
}

