#include "hdt_int32.h"
#include "hdt_blob.h"
#include "hank.h"
#include "support.h" /* rdtscl */
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

/* This file is compiled twice to create caseful and nocase data types.
 * The relevate macro is ACBM_CASEFUL. */

#define SIZE_T u_int32_t
#define OFF_T u_int32_t

static char rcsid[] = "$Id: hdt_blob.c,v 1.12 2002/01/21 06:01:14 tlewis Exp $";

/*
 * This implementation is based on the instructions in Gusfield,
 * _Algorithms on Strings, Trees, and Sequences_, pp 157-164.
 * There are two big differences between this and the silicondefense
 * implementation:
 * 1) I have tried to use variable names (i, ione, itwo, ithree, du)
 * that track to Gusfield's usage.
 * 2) I go left to right, as does Gusfield.  I hope at some point
 * in the future to enhance this implementation to allow it to cross
 * chunk boundaries, and ltr behaviour is crucial for that.
 */

#if 0
#define PRINTF(x...) printf(##x)
#define FPRINTF(x...) fprintf(##x)
#else
#define PRINTF(x...)
#define FPRINTF(x...)
#endif

#ifndef MIN
#define MIN(a,b) ((a<b)?a:b)
#endif

#ifndef MAX
#define MAX(c,d) ((c>d)?c:d)
#endif

#define NNULL(x) (x > 0 ? x : ~0)

#define ACBM_REPORT_MARK    (1<<0)
#define ACBM_REPORT_GST     (1<<1)
#define ACBM_REPORT_ZV      (1<<2)
#define ACBM_REPORT_DU      (1<<3)
#define ACBM_REPORT_DPRIMEV (1<<4)

#define DUMP_ACBM_TREE dump_acbm_tree

extern struct memcon *mc;

static int ic;

/* 1) The layout of this data structure tracks Gusfield's description
 * of the (lambda) tree, p163.
 * 2) This data structure represents both an edge with a label and
 * the node immediately underneath it in the suffix tree.  The root
 * has an empty label.
 * 3) Instead of having a jump table, this is a tree, with the children
 * in an ordered, linked list.  Terminal nodes have child==NULL.
 * This should save a significant amount of space without much cost,
 * since the pattern space doesn't look especially bushy to me.  Later,
 * if need be, that can be changed to a hash table, which can work well
 * with well less than 256 entries since the active set of the alphabet is
 * very local (e.g., a-zA-Z work very well in a small hash table.)
 */
struct acbm_node
{
	u_int8_t cmask;
  struct acbm_node **child;	/* This is the first of the children. */
  struct acbm_node *next;	/* This is the next child of the parent. */

  /* This is the content of this edge, along with its length */
  unsigned char *label;		/* N.b., binary data, not null-terminated */
  SIZE_T len;

  unsigned int pattern;			/* pattern number */
  OFF_T z;			/* suffix starting position */

  unsigned int marked:1;	/* step 2 */
  OFF_T zv;			/* step 3 */
  OFF_T du;			/* step 4 */
  OFF_T dprimev;		/* step 5 */
};

struct blob_match
{
  unsigned char *pattern;
  SIZE_T len;
  prim_id prid;
  OFF_T offset;
  OFF_T depth;
};

struct hdt_blob_set
{
  int kids;
  struct blob_match patterns[1024];
  SIZE_T pmin;
  SIZE_T pmax;
  SIZE_T ptot;
  int pattern_cursor;
  int built;
  struct acbm_node *tree;
  SIZE_T *bad_char[256];
  struct memcon_context *c;
};

static hdt_set intsets[8192];
static struct hdt_blob_set *sets[16384];
static int hdt_blob_sets_offset = 1;

/* Identical to memcmp, except that the first string
 * is converted to lowercase before the comparison. */
static int
memcasecmp(void *s1, void *s2, SIZE_T n)
{
	unsigned int i;
	unsigned char c;
	unsigned char *s = s2;

	for(i=0;i<n;i++){
		c=tolower(((char*)s1)[i]);
		if(c < s[i]){
			PRINTF("MEMCASECMP: %c < %c\n", c, s[i]);
			return(-1);
		}
		if(c > s[i]){
			PRINTF("MEMCASECMP: %c > %c\n", c, s[i]);
			return(1);
		}
	}
	return(0);
}

static void
dump_table (struct hdt_blob_set *h)
{
  unsigned int i, j;
  int found;
  SIZE_T x = h->pmax;

  for (i = 0; i < 256; i++)
    {				/* for each character */
      found = 0;
    dump_replay:
      if (found || 1){
	PRINTF ("%3i (%c): ", i, isgraph(i) ? i : 88);
		}

      for (j = 1; j <= x; j++)
	{
	  if (found){
	    PRINTF ("%2i ", h->bad_char[i][x - j]);
	  } else if (h->bad_char[i][x - j])
	    {
	      found = 1;
	      goto dump_replay;
	    }
	}
      if (found){
	PRINTF ("\n");
		}
    }
}

static void
dump_acbm_node (FILE * f, struct acbm_node *a, int parent, u_int32_t flags)
{
  unsigned char buf[1024];
  int i, id = ic++;
  struct acbm_node *b;

	if(!a)
		return;

  memcpy (buf, a->label, MIN (a->len, sizeof (buf)));
  buf[MIN (a->len, sizeof (buf))] = 0;

  FPRINTF (f, "	%i [", id);

  if (flags & ACBM_REPORT_MARK && a->marked)
    {
      FPRINTF (f, "style=filled,");
    }

  FPRINTF (f, "label=\"");

  if (flags & ACBM_REPORT_GST)
    {
      FPRINTF (f, "pattern: %i\\nz: %i\\nlen: %i\\n--------\\n", a->pattern,
	       a->z, a->len);
    }
  if (flags & ACBM_REPORT_ZV)
    {
      FPRINTF (f, "zv: %i\\n", a->zv);
    }
  if (flags & ACBM_REPORT_DU)
    {
      FPRINTF (f, "du: %i\\n", a->du);
    }
  if (flags & ACBM_REPORT_DPRIMEV)
    {
      FPRINTF (f, "d'v: %i\\n", a->dprimev);
    }
  FPRINTF (f, "\"");

  FPRINTF (f, "];\n");

	if (parent != -1) {
		if (a->len)
			FPRINTF (f, "	%i -> %i [label=\"%10s\"];\n", parent, id, buf);
		else
			FPRINTF (f, "	%i -> %i [label=\"($)\"];\n", parent, id);
	}
	PRINTF ("Dumping kids on %i slots.\n", a->cmask + 1);
	for (i = 0; i < (a->cmask + 1); i++) {
		PRINTF ("Slot %i\n", i);
		for (b = a->child[i]; b; b = b->next)
			dump_acbm_node (f, b, id, flags);
	}

}

static void
dump_acbm_tree (struct hdt_blob_set *a, const char *file_name, u_int32_t flags)
{
  int i;
  FILE *f;

  /* f=popen("dot -Tps | gv -", "r"); */
  unlink (file_name);
  f = fopen (file_name, "w");
  ic = 1;
  PRINTF ("acbm tree dump:\n\n");
  FPRINTF (f, "digraph X {\n");

  FPRINTF (f, "	struct1 [shape=record,label=\"");

  FPRINTF (f, "{pattern #");

  for (i = 0; a->patterns[i].pattern; i++)
    FPRINTF (f, "| %i", i);

  FPRINTF (f, "}|{pattern");

  for (i = 0; a->patterns[i].pattern; i++)
    FPRINTF (f, "|%10s", a->patterns[i].pattern);

  FPRINTF (f, "}");

  FPRINTF (f, "\"];\n");

#if 0
  dump_acbm_node (f, a->tree, -1, flags);
#else
	/* disabled */
	FPRINTF (f, "title=\"ACBM graphs disabled; see file %s, line %i\"\n", __FILE__, __LINE__);
#endif

  FPRINTF (f, "}\n");
  fclose (f);
}

static struct acbm_node *
create_acbm_node (SIZE_T size, unsigned char *content, int pattern,
		  OFF_T start_pos, struct memcon_context *c, u_int8_t hash_size)
{
  struct acbm_node *r;

  if ((int) size < 0)
    {
      PRINTF ("Aiee!  Negative node size.\n");
      exit (-1);
    }

  if ((r = c->eng->alloc (c, sizeof (*r))) == NULL)
    exit (-1);
  memset (r, 0, sizeof (*r));
  r->pattern = pattern;
  r->len = size;
  r->z = start_pos;
  if (r->len)
    {
      if ((r->label = c->eng->alloc(c, r->len)) == NULL)
	exit (-1);
      memcpy (r->label, content, r->len);
    }

	if((hash_size < 1) || (hash_size > 255)){
		fprintf(stderr, "AIEE!	received illegal hash size, %i, "
			"for acbm node.	Aborting.\n", hash_size);
		exit(-1);
	}

	r->cmask=hash_size-1;
	if((r->child=c->eng->alloc (c, (hash_size*sizeof(void*))) ) == NULL)
		exit(-1);
	memset(r->child, 0, (hash_size*sizeof(void*)));

  return (r);
}

static int
second_phase (struct acbm_node *r)
{
  struct acbm_node *s;
  int i = 0;

  for (s = r->child[0]; s; s = s->next)
    if (second_phase (s))
      i = 1;

  if (r->z == 1)
    i = 1;

  return (r->marked = i);
}

static int
third_phase (struct acbm_node *r)
{
  struct acbm_node *s;
  int i = 0, j;

  if (r->child[0] == NULL)
    {
      i = r->z;			/* leaf */
    }
  else
    {
      for (s = r->child[0]; s; s = s->next)
	{
	  j = third_phase (s);
	  if (j > 1 && (i == 0 || j < i))
	    i = j;
	}
    }

  if (r->marked && i > 1)
    r->zv = i;

  return (i);
}

static int
fourth_phase (struct acbm_node *r)
{
  struct acbm_node *s;
  int i;

  if (!r->child[0] && !r->len)	/* root is empty but does not count */
    return (r->z);

  for (s = r->child[0]; s; s = s->next)
    if ((i = fourth_phase (s)))
      r->du = i;

  return (0);
}

static void
fifth_phase (struct acbm_node *r, unsigned int value)
{
  struct acbm_node *s;
  unsigned int i = value;

  if (!i || (r->du && (r->du < i)))
    i = r->du;

  r->dprimev = i;

  for (s = r->child[0]; s; s = s->next)
    fifth_phase (s, i);
}

static void
sixth_phase (struct acbm_node *r)
{
  struct acbm_node *s, *t = NULL, *u;

  for (s = r->child[0]; s;)
    {
      if (s->marked)
	{			/* This node is safe; look at kids */
	  sixth_phase (s);
	  t = s;
	  s = s->next;
	}
      else
	{			/* Destroy this subtree */
	  if (t)
	    t->next = s->next;
	  else
	    r->child[0] = s->next;

	  u = s->next;
	  s = u;
	}
    }
  if (t == NULL)
    r->child[0] = NULL;

}

/* case */
#define CCONV(x) x
#define MEMCMP(x...) memcmp(##x)
#define ACBM_MATCH "acbm_case_match"
#define TYPE(x) hdt_cblob ## x
#include "hdt_blob_impl.c"

#undef CCONV
#undef MEMCMP
#undef ACBM_MATCH
#undef TYPE

/* nocase */
#define CCONV(x) (unsigned char)tolower(x)
#define MEMCMP(x...) memcasecmp(##x)
#define ACBM_MATCH "acbm_nocase_match"
#define TYPE(x) hdt_ncblob ## x
#define ACBM_CASELESS
#include "hdt_blob_impl.c"

#undef CCONV
#undef MEMCMP
#undef ACBM_MATCH
#undef TYPE

hdt_set
hdt_blob_create (void)
{
  hdt_set r = hdt_blob_sets_offset;
  struct memcon_context *c;
  struct hdt_blob_set *h;
  int i;

  for (i = 0; i < 2; i++) {
    if ((unsigned int) r >= sizeof (sets))	/* out of room! */
      return (-1);

    c = mc->create (NULL, "hdt_blob");
    if (c == NULL)
      {
	fprintf (stderr, "Aiee!  memcon->create returned NULL!  Exiting.\n");
	exit (-1);
      }

    if ((h = c->eng->alloc (c, sizeof (*h))) == NULL)
      {
	fprintf (stderr, "Aiee!  memcon->alloc returned NULL!  Exiting.\n");
	exit (-1);
      }
    memset (h, 0, sizeof (*h));

    h->c = c;

    sets[r + i] = h;
  }

	hdt_blob_sets_offset += 2;
  return (r);
}

void
hdt_blob_destroy (hdt_set set)
{
  sets[set]->c->eng->destroy (sets[set]->c);
  sets[set] = NULL;
}

hdt_prim
hdt_blob_add (hdt_set set, prim_id prid, char *op, char *arg)
{
	PRINTF("Adding blob match - op: %s, arg: %s\n", op, arg);
	if (!strcmp (op, "acbm_nocase_match")){
		return(hdt_ncblob_add(set, prid, arg));
	} else if (!strcmp (op, "acbm_case_match")){
		return(hdt_cblob_add(set+1, prid, arg));
	} else {
		if(!intsets[set])
			intsets[set] = hdt_int32_create();
		return(hdt_int32_add(intsets[set], prid, op, arg));
	}

	FPRINTF(stderr, "HDT_BLOB: unrecognized operation \"%s\".\n", op);
	return(0);
}

void
hdt_blob_del(hdt_set set, hdt_prim prim)
{
  /* UNSUPPORTED */
}

int
hdt_blob_match(hdt_set set, hdt_prim * prims, struct blob *val,
		unsigned char *rbuf, char newbits)
{
	int i;
	unsigned long t1, t2, t3, t4;

	printf("** HDT_BLOB: Starting\n");

	rdtscl(t1);

	if(sets[set]->kids && (i=hdt_ncblob_match(set, prims, val, rbuf, newbits))){
		FPRINTF(stderr, "AIEE!  hdt_ncblob_match() returned %i!  Aborting.\n", i);
		return(i);
	}

	rdtscl(t2);
	if(sets[set+1]->kids && (i=hdt_cblob_match(set+1, prims, val, rbuf, newbits))){
		FPRINTF(stderr, "AIEE!  hdt_cblob_match() returned %i!  Aborting.\n", i);
		return(i);
	}

	rdtscl(t3);
	if(intsets[set]){
		PRINTF("Running integer ops on blob length of \"%i\".\n", val->len);
		if(i=hdt_int32_match(intsets[set], prims, val->len, rbuf, newbits)){
			FPRINTF(stderr, "AIEE!  hdt_int32_match() returned %i!  Aborting.\n", i);
			return(i);
		}
	}
	rdtscl(t4);

	printf("nocase: %i bytes, %ul cycles\n", val->len, t2-t1);
	printf("case: %i bytes, %ul cycles\n", val->len, t3-t2);
	printf("hdt_blob/int32: %i bytes, %ul cycles\n", val->len, t4-t3);
	printf("** HDT_BLOB: Overall hdt_blob processing: %ul cycles.\n", t4-t1);
	fflush(stdout);

	return(0);
}

void
build_blobs(void)
{
	int i;

	for(i=0;i<16384;i += 2){
		if(!sets[i])
			break;
		if (sets[i]->built == 0)
			hdt_ncblob_build_acbm (sets[i]);
		if (sets[i+1]->built == 0)
			hdt_cblob_build_acbm (sets[i+1]);
	}
}

char *
hdt_blob_report(struct blob *val, struct memcon_context *c)
{
  return (NULL);
}
