/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

/* memcon.c: memory context implementation*/

#include "hank.h"
#include "memcon.h"
#include "smodule.h"
#include "types.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define SIZE_T u_int32_t

#define PRINTF(x...)

static char rcsid[] = "$Id: memcon_malloc.c,v 1.10 2002/01/03 22:02:28 tlewis Exp $";

/* CONSTANTS */
#define NSLOTS 65535
#define MCVMAJ 1
#define MCVMIN 1

/* TYPE DEFINITIONS */
struct malloc_context {
	struct memcon_context c;
	void** allocations;
};

/* PROTOTYPES */
static struct memcon_context *my_create(struct memcon_context *context, const char* owner);
static void my_destroy(struct memcon_context *c);
static void *my_alloc(struct memcon_context *c, SIZE_T size);
static void my_free(struct memcon_context *c, void* chunk);
static void my_dump_contexts(struct memcon_context *c, void (callback)(struct memcon_context*));

/* memcon */
struct memcon mallocmc = {
	{
		"memcon",
		"malloc",
		MCVMAJ,
		MCVMIN,
		&mallocmc
	},
	&my_create,
	&my_destroy,
	&my_alloc,
	&my_free,
	&my_dump_contexts
};

/* PRIVATE VARIABLES */
static struct malloc_context root = 
	{{NULL, NULL, NULL, NULL, &mallocmc, 0, NULL, 0}, NULL};
static int id=0;

/* PRIVATE FUNCTIONS */

static void
malloc_failed(int location)
{
	fprintf(stderr, "MALLLOC FAILED at %s-%s:%i; exiting.\n", 
		__DATE__, __FILE__, location);
	exit(-1);
}

/* really zap a context; keyed by actual context, not id */
static void 
memcon_zap(struct malloc_context *x)
{
	struct memcon_context *c=&(x->c), *d, *e;
	void **v;
	int i;

	/* first, recurse */
	for(d=c->child;d;){
		e=d->next;
		memcon_zap((struct malloc_context *)d);
		d=e;
	}

	/* second, take it out of the lists */
	if(c->next)
		c->next->prev=c->prev;
	if(c->prev)
		c->prev->next=c->next;
	if(c->parent)
		if(c->parent->child==c)
			c->parent->child=c->next;

	/* third, free the allocations */
   for(i=0;i<NSLOTS;i++){
      v=&(x->allocations[i]);
		PRINTF("memcon:alloc(): freeing allocation %i at %p.\n", i, *v);
		free(*v);
	}

	/* fourth, free the structure */
	free(x->allocations);
	free(x);
}

static void
really_dump_contexts(struct memcon_context *c, void (*callback)(struct memcon_context*))
{
	struct memcon_context *d;

	(*callback)(c);
	for(d=c->child;d;d=d->next)
		really_dump_contexts(d, callback);
}

/* PUBLIC FUNCTIONS */

/* create context */
static struct memcon_context *
my_create(struct memcon_context *context, const char* owner)
{
	struct malloc_context *x;
	struct memcon_context *c, *d, *parent;

/*
	if(context)
		PRINTF("my_create(): %s\n", context->owner);
	else
		PRINTF("my_create(): (NULL)\n");
*/

	if(context==NULL){
		parent=&(root.c);
	} else {
		parent=context;
	}

	x=malloc(sizeof(struct malloc_context));
	if(x==NULL)
		malloc_failed(__LINE__);
	memset(x, 0, sizeof(struct malloc_context));

	c=&(x->c);
	c->id=id++;
	c->owner=owner;
	c->parent=parent;
	c->eng=&mallocmc;
	x->allocations=malloc(NSLOTS*(sizeof(void*)));
	if(x->allocations==NULL)
		malloc_failed(__LINE__);
	memset(x->allocations, 0, (NSLOTS*(sizeof(void*))));
	d=parent->child;
	parent->child=c;
	c->next=d;
	if(d)
		d->prev=c;

	return(c);
}

/* destroy context */
static void 
my_destroy(struct memcon_context *c)
{
	PRINTF("memcon->destroy() called; size %i.\n", c->size);
	memcon_zap((struct malloc_context *)c);
}

/* alloc */
static void* 
my_alloc(struct memcon_context *m, SIZE_T size)
{
   int i;
   void** v;
	struct malloc_context *n=(struct malloc_context*)m;


   if(m==NULL){
		printf("Aiee!  Received null context!  Returning null.\n");
      return(NULL);
	}

   for(i=0;i<NSLOTS;i++){
      v=&(n->allocations[i]);
      if(*v==NULL){
         *v=malloc(size);
         memset(*v, 0, size);
			m->size+=size;
         if(*v==NULL)
            malloc_failed(__LINE__);
			PRINTF("memcon:alloc(): allocated %i bytes at %i(%p).\n", size, i, *v);
         return(*v);
		}
   }
	printf("Aiee!  Apparently ran out of slots in memcon_alloc.  Returning null.\n");
   return(NULL);
}

/* free */
static void 
my_free(struct memcon_context *c, void* chunk)
{
	if(c==chunk)
		return;
	return;
}

void
my_dump_contexts(struct memcon_context *c, void (callback)(struct memcon_context*))
{
	really_dump_contexts(c, callback);
}
