/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

/* memcon.c: memory context implementation*/

#include "hank.h"
#include "smodule.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static char rcsid[] = "$Id: memcon_malloc.orig.c,v 1.5 2001/11/10 01:45:33 tlewis Exp $";

/* CONSTANTS */
#define NSLOTS 256
#define MCVMAJ 1
#define MCVMIN 1

/* TYPE DEFINITIONS */
/* This is a redefinition from the header file */
struct memcon_context {
	struct memcon_context* parent;
	struct memcon_context* child;
	struct memcon_context* prev;
	struct memcon_context* next;
	int id;
	const char* owner;
	size_t size;
	void** allocations;
} memcon_context;

typedef int memcon_cid;

/* PROTOTYPES */
static memcon_cid my_create(memcon_cid context, const char* owner);
static void my_destroy(memcon_cid c);
static void *my_alloc(memcon_cid c, size_t size);
static void my_free(memcon_cid c, void* chunk);
static void my_dump_contexts(memcon_cid c, void (callback)(struct memcon_context*));

/* memcon */
struct memcon {
	struct smodule s;
	memcon_cid (*create)(memcon_cid context, const char* owner);
	void (*destroy)(memcon_cid c);
	void* (*alloc)(memcon_cid c, size_t size);
	void (*free)(memcon_cid c, void* chunk);
	void (*dump_contexts)(memcon_cid c, void (*callback)(struct memcon_context*));
} mc = {
	{
		"memcon",
		"malloc",
		MCVMAJ,
		MCVMIN,
		&mc
	},
	&my_create,
	&my_destroy,
	&my_alloc,
	&my_free,
	&my_dump_contexts
};

/* PRIVATE VARIABLES */
static memcon_cid cid=1;
static struct memcon_context root = 
	{NULL, NULL, NULL, NULL, 0, "", 0, NULL};

/* PRIVATE FUNCTIONS */

static void
malloc_failed(int location)
{
	fprintf(stderr, "MALLLOC FAILED at %s-%s:%i; exiting.\n", 
		__DATE__, __FILE__, location);
	exit(-1);
}

static struct memcon_context* 
find_cont_by_id(struct memcon_context* c, memcon_cid i)
{
	struct memcon_context *m, *n;

	for(m=c;m;m=m->next){
		/* first check this node */
		if(m->id==i)
			return(m);
		/* second, recurse on the child */
		for(m=m->child;m;m=m->next){
			if((n=find_cont_by_id(m, i))!=NULL)
				return(n);
		}
	}
	return(NULL);
}

/* really zap a context; keyed by actual context, not id */
static void 
memcon_zap(struct memcon_context* c)
{
	struct memcon_context *d, *e;
	void **v;

	/* first, recurse */
	for(d=c->child;d;){
		e=d->next;
		memcon_zap(d);
		d=e;
	}

	/* second, take it out of the lists */
	if(c->next)
		c->next->prev=c->prev;
	if(c->prev)
		c->prev->next=c->next;
	if(c->parent)
		if(c->parent->child==c)
			c->parent->child=c->next;

	/* third, free the allocations */
	for(v=c->allocations;*v!=NULL;v++)
		free(*v);

	/* fourth, free the structure */
	free(c->allocations);
	free(c);
}

static void
really_dump_contexts(struct memcon_context *c, void (*callback)(struct memcon_context*))
{
	struct memcon_context *d;

	(*callback)(c);
	for(d=c->child;d;d=d->next)
		really_dump_contexts(d, callback);
}

/* PUBLIC FUNCTIONS */

/* create context */
static memcon_cid 
my_create(memcon_cid context, const char* owner)
{
	memcon_cid i=cid++;
	struct memcon_context *c, *d, *parent;

	if(context==0){
		parent=&root;
	} else {
		parent=find_cont_by_id(&root, context);
		if(parent==NULL)
			return(-1);
	}

	c=malloc(sizeof(struct memcon_context));
	if(c==NULL)
		malloc_failed(__LINE__);
	memset(c, 0, sizeof(struct memcon_context));


	c->id=i;
	c->owner=owner;
	c->parent=parent;
	c->allocations=malloc(NSLOTS*(sizeof(void*)));
	if(c->allocations==NULL)
		malloc_failed(__LINE__);
	memset(c->allocations, 0, (NSLOTS*(sizeof(void*))));
	d=parent->child;
	parent->child=c;
	c->next=d;
	if(d)
		d->prev=c;

	return(i);
}

/* destroy context */
static void 
my_destroy(memcon_cid c)
{
	memcon_zap(find_cont_by_id(&root, c));
}

/* alloc */
static void* 
my_alloc(memcon_cid m, size_t size)
{
	int i;
	void** v;
	struct memcon_context *c = find_cont_by_id(&root, m);

	if(c==NULL)
		return(NULL);

	for(i=0;i<NSLOTS;i++){
		v=&(c->allocations[i]);
		if(*v==NULL){
			*v=malloc(size);
			memset(*v, 0, size);
			if(*v==NULL)
				malloc_failed(__LINE__);
			return(*v);
		}
	}
	return(NULL);
}

/* free */
static void 
my_free(memcon_cid c, void* chunk)
{
	if(c!=1||chunk==NULL)
		return;
	return;
}

void
my_dump_contexts(memcon_cid c, void (callback)(struct memcon_context*))
{
	really_dump_contexts(find_cont_by_id(&root, c), callback);
}
