/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

/* memcon.c: memory context implementation*/

/* Dude, this is gonna rock!  Actually returning memory to the system! */

#include "hank.h"
#include "memcon.h"
#include "smodule.h"
#include "types.h"
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>

#define SIZE_T u_int32_t

static char rcsid[] = "$Id: memcon_mmap.c,v 1.8 2002/01/03 22:02:28 tlewis Exp $";

#define PRINTF(x...)

/* CONSTANTS */
#define NSLOTS 65535
#define MCVMAJ 1
#define MCVMIN 1
#define PAGESIZE 4096

struct mapping {
	unsigned char *map;
	SIZE_T len;
	SIZE_T consumed;
	struct mapping *next;
};

/* TYPE DEFINITIONS */
struct mmap_context {
	struct memcon_context c;
	struct mapping *mappings;
};

static void *free_pages[32];
static int get_cur=0, put_cur=0, npages=0;

static int
put_page(void *page)
{
	if(free_pages[put_cur%32]==NULL){
		/* there's a slot */
		PRINTF("Putting page (slot %i, npages %i, addr %p)\n", put_cur%32, npages, page);
		free_pages[put_cur%32]=page;
		put_cur++;
		npages++;
		return(1);
	}
	return(0);
}

static void *
get_page(void)
{
	if(free_pages[get_cur%32]!=NULL){
		void *v;
		/* there's a slot */
		v=free_pages[get_cur%32];
		PRINTF("Getting page (slot %i, npages %i, addr %p)\n", get_cur%32, npages, v);
		get_cur++;
		npages--;
		free_pages[get_cur%32]=NULL;
		return(v);
	}
	return(NULL);
}


/* PROTOTYPES */
static struct memcon_context *my_create(struct memcon_context *context, const char* owner);
static void my_destroy(struct memcon_context *c);
static void *my_alloc(struct memcon_context *c, SIZE_T size);
static void my_free(struct memcon_context *c, void* chunk);
static void my_dump_contexts(struct memcon_context *c, void (callback)(struct memcon_context*));

/* memcon */
struct memcon mmapmc = {
	{
		"memcon",
		"mmap",
		MCVMAJ,
		MCVMIN,
		&mmapmc
	},
	&my_create,
	&my_destroy,
	&my_alloc,
	&my_free,
	&my_dump_contexts
};

/* PRIVATE VARIABLES */
static struct mmap_context root = 
	{{NULL, NULL, NULL, NULL, &mmapmc, 0, NULL, 0}, NULL};
static int id=0;
static int mmap_devzero = 0;

/* PRIVATE FUNCTIONS */

static void
mmap_failed(int location)
{
	fprintf(stderr, "MMAP FAILED at %s-%s:%i; exiting.\n", 
		__DATE__, __FILE__, location);
	exit(-1);
}

/* really zap a context; keyed by actual context, not id */
static void 
memcon_zap(struct mmap_context *x)
{
	struct memcon_context *c=&(x->c), *d, *e;
	struct mapping *u, *v;

	/* first, recurse */
	for(d=c->child;d;){
		e=d->next;
		memcon_zap((struct mmap_context *)d);
		d=e;
	}

	/* second, take it out of the lists */
	if(c->next)
		c->next->prev=c->prev;
	if(c->prev)
		c->prev->next=c->next;
	if(c->parent)
		if(c->parent->child==c)
			c->parent->child=c->next;

	/* third, free the allocations */
	for(v=x->mappings;v;){
		if(v->len==PAGESIZE)
			if(put_page(v->map))
				goto bar;
		PRINTF("MEMCON_MMAP: munmap()'ing %i bytes at %p\n", v->len, v->map);
		munmap(v->map, v->len);
bar:
		u=v;
		v=v->next;
		free(u);
	}

	/* fourth, free the structure */
	free(x);
}

static void
really_dump_contexts(struct memcon_context *c, void (*callback)(struct memcon_context*))
{
	struct memcon_context *d;

	(*callback)(c);
	for(d=c->child;d;d=d->next)
		really_dump_contexts(d, callback);
}

static void *
dnmmap(SIZE_T len, int prot, int flags)
{
	void *v;

	if(mmap_devzero==0)
		if((mmap_devzero=open("/dev/zero", O_RDWR))==-1)
			exit(-1);

	if((v=mmap(NULL, len, prot, flags, mmap_devzero, 0))==MAP_FAILED){
		fprintf(stderr, "AIEE!  mmap of /dev/zero failed!  Exiting.\n");
		perror("errno:");
		exit(-1);
	}
	PRINTF("MEMCON_MMAP: mmap()'d %i bytes at %p.\n", len, v);

	return(v);
}

/* PUBLIC FUNCTIONS */

/* create context */
static struct memcon_context *
my_create(struct memcon_context *context, const char* owner)
{
	struct mmap_context *x;
	struct memcon_context *c, *d, *parent;

	if(context==NULL){
		parent=&(root.c);
	} else {
		parent=context;
	}

	x=malloc(sizeof(struct mmap_context));
	if(x==NULL)
		mmap_failed(__LINE__);
	memset(x, 0, sizeof(struct mmap_context));

	c=&(x->c); /* same as "=x;", but more typer-friendly */
	c->id=id++;
	c->owner=owner;
	c->parent=parent;
	c->eng=&mmapmc;
	x->mappings=NULL;
	d=parent->child;
	parent->child=c;
	c->next=d;
	if(d)
		d->prev=c;

	PRINTF("mmap: Returning memcon context at %p.\n", c);
	return(c);
}

/* destroy context */
static void 
my_destroy(struct memcon_context *c)
{
	PRINTF("memcon->destroy() called; size %i.\n", c->size);
	memcon_zap((struct mmap_context *)c);
}

/* alloc */
static void* 
my_alloc(struct memcon_context *m, SIZE_T size)
{
	struct mmap_context *n=(struct mmap_context*)m;
	struct mapping *g;
	SIZE_T z;
	void *v;

   if(m==NULL){
		printf("Aiee!  Received null context!  Returning null.\n");
      return(NULL);
	}

	g=n->mappings;

	/* 32-bit aligned */
	z = size + (4 - ((size%4))%4);

	if (z > PAGESIZE) {
		/* create a special mapping and tuck */
		if((g=malloc(sizeof(*g)))==NULL)
			exit(-1);

		/* create */
		g->len=z;
		g->consumed=z;
		g->map=dnmmap(z, PROT_READ|PROT_WRITE, MAP_PRIVATE);

		/* tuck */
		if(n->mappings){
			g->next=n->mappings->next;
			n->mappings->next=g;
		} else {
			g->next=NULL;
			n->mappings=g;
		}

/*
		for(i=0;i<g->len;i++)
			(char *)g->map[i]=3;
*/

		return(g->map);

	} else if ( (g==NULL) || (z > (g->len - g->consumed)) ){
		/* create a new mapping */
		if((g=malloc(sizeof(*g)))==NULL)
			exit(-1);

		/* create */
		g->len=PAGESIZE;
		g->consumed=z;
		if((g->map=get_page())==NULL)
			g->map=dnmmap(PAGESIZE, PROT_READ|PROT_WRITE, MAP_PRIVATE);

		/* stick*/
		if(n->mappings){
			g->next=n->mappings;
			n->mappings=g;
		} else {
			g->next=NULL;
			n->mappings=g;
		}

/*
		for(i=0;i<g->len;i++)
			(char *)g->map[i]=3;
*/

		return(g->map);

	} else {
		/* there's space */
		v=&g->map[g->consumed];
		g->consumed += z;
/*
		for(i=0;i<z;i++)
			((char *)v)[i]=3;
*/
		return(v);
	}

	printf("AIEE!  mmap_alloc failed!  Returning NULL.\n");
   return(NULL);
}

/* free */
static void 
my_free(struct memcon_context *c, void* chunk)
{
	if(c==chunk)
		return;
	return;
}

void
my_dump_contexts(struct memcon_context *c, void (callback)(struct memcon_context*))
{
	really_dump_contexts(c, callback);
}
