#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif /*_GNU_SOURCE*/

#include "hank.h"
#include "smodule.h"
#include "types.h"
#include <dirent.h>
#include <dlfcn.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

static char rcsid[] = "$Id: smodule_simple.c,v 1.8 2002/01/03 22:02:28 tlewis Exp $";

/* Number of hank module collections */
#define NSMCOLS 256
#define SMNS 256

struct sminternal {
	struct smodule** s;
	void* m;
	int ref_count;
	struct sminternal *p;
	struct sminternal *n;
};

struct smcol_tr {
	int in_use;
	char** dirs;
	struct smodule*** statics;
	char* mod_name;
	unsigned int vmaj;
	unsigned int vmin;
	struct sminternal *aa; /* already activated */
};

static smcol_t scan_collection(const char* mod_name, struct smodule*** statics, 
	const char** dirs, unsigned int vmaj, unsigned int vmin);
static void release_collection(smcol_t c);
static struct smodule* get_smodule(smcol_t c, const char* name);
static void release_smodule(smcol_t c, struct smodule* mod);

struct smodule_juggler sj = {
	&scan_collection,
	&release_collection,
	&get_smodule,
	&release_smodule
};

static struct smcol_tr collections[NSMCOLS];

/* returns 1 if they're ok, 0 if not */
static int
confirm_versions(struct smodule *s, struct smcol_tr *t)
{
	if(t->vmaj!=s->vmaj)
		return(0);
	
	if(t->vmin > s->vmin)
		return(0);
	
	return(1);
}

static char** 
dup_paths(const char** p)
{
	char** r;
	char* b;
	int i;

	if(p==NULL)
		return(NULL);

	for(i=0;p[i]!=NULL;i++){
	}

	/* fprintf(stderr, "Path has %i components.\n", i); */
	r=(char**)malloc((i+1)*(sizeof(char *)));
	if(r==NULL){
		fprintf(stderr, "Malloc failed!\n");
		return(NULL);
	}

	for(;i>-1;i--){
		if(p[i]==NULL){
			r[i]=NULL;
		} else {
			b=strdup(p[i]);
			if(b==NULL){
				/* FIXME: free previous mallocs */
				fprintf(stderr, "Malloc failed!\n");
				return(NULL);
			}
			r[i]=b;
		}
	}
	return(r);
}

static smcol_t
scan_collection(const char* mod_name, struct smodule*** statics, 
	const char** dirs, unsigned int vmaj, unsigned int vmin)
{
	int i;

	for(i=0;i<NSMCOLS && collections[i].in_use;i++){
	}

	if(i==NSMCOLS){
		fprintf(stderr, "ERROR: smodule system: maximum number of collections (%i) in use.\n", i);
		errno=ENOMEM;
		return(-1);
	}
	collections[i].in_use=1;
	collections[i].vmaj=vmaj;
	collections[i].vmin=vmin;
	collections[i].statics=statics;
	collections[i].dirs=dup_paths(dirs);
	if(mod_name==NULL){
		collections[i].mod_name=strdup("smodule");
	} else {
		collections[i].mod_name=strdup(mod_name);
	}
	if(collections[i].mod_name==NULL){
		return(-1);
	}
	return(i);
}

static void 
release_collection(smcol_t c)
{
	if(c>NSMCOLS){
		errno=EINVAL;
		return;
	}

	/* FIXME: free dirs */
	collections[c].in_use=0;
	free(collections[c].mod_name);
}

static struct smodule* 
get_smodule(smcol_t c, const char* name)
{
	int i=0, j, k;
	struct smodule*** a;
	struct smodule** s;
	struct smodule* t;
	struct sminternal* n;
	char* b;
	DIR* d;
	struct dirent* de;
	char namebuf[256];
	struct stat fsb;
	void* dlhandle;

	if(c>NSMCOLS){
		errno=EINVAL;
		return(NULL);;
	}

	/* First, scan already-activated modules */
	for(n=(collections[c]).aa;n;n=n->n){
		for(i=0;n->s[i];i++){
			t=n->s[i];
			if((!(strcmp(name, t->name))) && confirm_versions(t, &collections[c])){
				fprintf(stderr, "smodule: found already-activated module \"%s\" as \"%s\"!\n", name, t->name);
				n->ref_count++;
				return(t);
			}
		}
	}

	/* Second, scan dynamic modules in passed directory order */
	if((collections[c].aa != NULL) && (collections[c].dirs[i] != NULL)){
		for(i=0;collections[c].dirs[i];i++){
			b=(collections[c].dirs)[i];
			if((d=opendir(b))==NULL){
				fprintf(stderr, "ERROR: smodule: can't opendir \"%s\": %s\n", b, strerror(errno));
			} else {
				/* fprintf(stderr, "smodule: opendir of \"%s\" went fine.\n", b); */
			}
			while(d&&(de=readdir(d))!=NULL){
				if((j=snprintf(namebuf, sizeof(namebuf), "%s/%s", b, de->d_name))==-1){
					fprintf(stderr, "smodule: snprintf failed(%i); skipping %s\n", j, de->d_name);
				} else if ((stat(namebuf, &fsb))||(!(fsb.st_mode&S_IFREG))) {
/* fprintf(stderr, "Direntry %s not openable; skipping.\n", namebuf); */
				} else if((dlhandle=dlopen(namebuf, RTLD_LAZY))==NULL){
					fprintf(stderr, "dlopen on %s failed.\n", namebuf);
				} else if((s=(struct smodule**)dlsym(dlhandle, collections[c].mod_name))==NULL){
					fprintf(stderr, "dlsym on %s failed.\n", de->d_name); 
					dlclose(dlhandle);
				} else {
					for(k=0;s[k]!=NULL;k++){
						t=s[k];
						if((!(strcmp(name, t->name))) && confirm_versions(t, &collections[c])){
							/* found it */
							if((n=malloc(sizeof(struct sminternal)))==NULL){
								fprintf(stderr, "smodule: Can't alocate memory!\n");
								return(NULL);
							}
							n->s=s;
							n->ref_count=1;
							n->p=NULL;
							n->n=(collections[c]).aa;
							n->m=dlhandle;
							if((collections[c]).aa){
								(collections[c]).aa->p=n;
							}
							(collections[c]).aa=n;
							fprintf(stderr, "smodule: found shared module!\n");
							return(t);
						} else {
							/* fprintf(stderr, "%s does not match %s; skipping.\n", t->name, name); */
							dlclose(dlhandle);
						}
					}
				}
			}
			closedir(d);
		}
	}

	/* Third, scan static modules */
	if((a=collections[c].statics)!=NULL){
		for(i=0;a[i]!=NULL;i++){
			s=a[i];
			for(j=0;s[j]!=NULL;j++){
				t=s[j];
				if((!(strcmp(name, t->name))) && confirm_versions(t, &collections[c])){
					/* fprintf(stderr, "smodule: found static module!\n"); */
					return(t);
				}
				/* fprintf(stderr, "smodule \"%s\" not match \"%s\".\n", name, t->name); */
			}
		}
	}

	/* fprintf(stderr, "smodule: Default returning NULL.\n"); */
	errno=ENOENT;
	return(NULL);
}

static void 
release_smodule(smcol_t c, struct smodule* mod)
{
	struct sminternal* n;
	struct smodule* s;
	int i;

	if(c>NSMCOLS){
		errno=EINVAL;
		return;
	}

	for(n=(collections[c]).aa;n;n=n->n){
		for(i=0;n->s[i];i++){
			s=n->s[i];
			if(s==mod){
				/* match! */
				/* fprintf(stderr, "Freeing module \"%s\".\n", mod->name); */
				if(!(--(n->ref_count))){
					if(n->n)
						n->n->p=n->p;
					if(n->p)
						n->p->n=n->n;
					dlclose(n->m);
					free(n);
				}
			}
		}
	}
}

